from datetime import datetime, timedelta

week = ["Sunday", "Monday", "Tuesday",
        "Wednesday", "Thursday", "Friday", "Saturday"]
day_name_map = {
    "Sunday": "Sun",
    "Monday": "Mon",
    "Tuesday": "Tue",
    "Wednesday": "Wed",
    "Thursday": "Thur",
    "Friday": "Fri",
    "Saturday": "Sat"
}
keys = [
    "frequency", "interval", "time",
    "weekDays", "monthDays",
    "hours", "minutes"
]
start_date = datetime(datetime.utcnow().year, 1, 1, 0, 0)
start_time_fmt = "%Y-%m-%dT%H:%M:%SZ"


def create_schedule_id(config_obj):
    entries = {k: config_obj.get(k) for k in keys}

    if entries["weekDays"]:
        entries["weekDays"] = \
            tuple([w for w in week if w in entries["weekDays"]])
    elif entries["monthDays"]:
        entries["monthDays"] = tuple(sorted(entries["monthDays"]))
    elif entries["frequency"] is None:
        entries["frequency"] = "Day"

    if entries["frequency"] is None:
        if entries["hours"] is None:
            entries["hours"] = [0]
        if entries["minutes"] is None:
            entries["minutes"] = [0]
    elif entries["minutes"] is None and entries["hours"] is not None:
        entries["minutes"] = [0]

    if entries["hours"] is not None:
        entries["hours"] = tuple(sorted(entries["hours"]))
    if entries["minutes"] is not None:
        entries["minutes"] = tuple(sorted(entries["minutes"]))

    return tuple(entries[k] for k in keys)


def trigger_name(frequency, interval, time,
                 weekDays, monthDays,
                 hours, minutes):

    def create_times_list():
        return " ".join([
            str(h).zfill(2) + str(m).zfill(2)
            for h in hours
            for m in minutes
        ])

    if frequency and interval:
        trigger_name = f"Every {interval} {frequency}"
        if interval > 1:
            trigger_name += "s"
        return trigger_name
    elif weekDays:
        return " - ".join([
            "Each Week",
            " ".join([day_name_map[day] for day in weekDays]),
            create_times_list()
        ])
    elif monthDays:
        return " - ".join([
            "Each Month",
            "Days " + " ".join([str(d) for d in sorted(monthDays)]),
            create_times_list()
        ])
    elif time:
        return "Daily - " + time.replace(":", "")
    else:
        return "Daily - " + create_times_list()


def create_recurrence_object(frequency, interval, time,
                             weekDays, monthDays,
                             hours, minutes):

    recurr_obj = {
        "interval": 1,
        "startTime": start_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
        "timeZone": "UTC"
    }

    if weekDays:
        return {
            **recurr_obj,
            "frequency": "Week",
            "schedule": {
                "hours": list(hours),
                "minutes": list(minutes),
                "weekDays": list(weekDays)
            }
        }
    elif monthDays:
        return {
            **recurr_obj,
            "frequency": "Month",
            "schedule": {
                "hours": list(hours),
                "minutes": list(minutes),
                "monthDays": list(monthDays)
            }
        }
    elif frequency in (None, "Day"):
        obj = {
            **recurr_obj,
            "frequency": "Day"
        }

        if time is not None:
            hour, minute = time.split(":")
        else:
            hour, minute = hours[0], minutes[0]
            obj["schedule"] = {
                "hours": list(hours),
                "minutes": list(minutes)
            }
        obj["startTime"] = (
                start_date + timedelta(hours=int(hour), minutes=int(minute))
            ).strftime(start_time_fmt)

        return obj
    else:
        recurr_obj["frequency"] = frequency
        if interval:
            recurr_obj["interval"] = interval

        return recurr_obj



def filter_for_files(output_activity):
    dep_name = output_activity["name"]
    return {
        "name": f"Only files from {dep_name}",
        "type": "Filter",
        "dependsOn": [
            {
                "activity": dep_name,
                "dependencyConditions": [
                    "Succeeded"
                ]
            }
        ],
        "userProperties": [],
        "typeProperties": {
            "items": {
                "value": f"@activity('{dep_name}').output.childItems",
                "type": "Expression"
            },
            "condition": {
                "value": "@equals(item().type, 'File')",
                "type": "Expression"
            }
        }
    }



from copy import deepcopy
import json
from os import listdir, makedirs, path, remove
from re import sub

from .defaults import default_annotations
from .schedule import create_schedule_id, create_recurrence_object, \
    trigger_name
from .sftp import FTPPipeline, SFTPPipeline
from .templates.dataset import all_data_sets
from .templates.linked_service import all_linked_services


class CreateDataFactoryObjects:

    connection_types = {
        "ftp": FTPPipeline,
        "sftp": SFTPPipeline
    }

    base_integration_runtime = "AutoResolveIntegrationRuntime"

    def __init__(self, config_folder=None, generated_folder=None,
                 overwrite=False, remove_not_generated=False):
        self.config_folder = config_folder or "configs"
        self.generated_folder = generated_folder or "generated"
        self.overwrite = overwrite
        self.remove_not_generated = remove_not_generated

        self.shir_folder = f"{self.generated_folder}/integrationRuntime"
        self.linked_service_folder = f"{self.generated_folder}/linkedService"
        self.data_set_folder = f"{self.generated_folder}/dataset"
        self.pipeline_folder = f"{self.generated_folder}/pipeline"
        self.trigger_folder = f"{self.generated_folder}/trigger"

        self.check_folders()

        self.all_linked_services = set()
        self.all_data_sets = set()

        self.all_config_jsons = []
        self.all_self_hosted_integration_runtimes = {}
        self.all_linked_service_jsons = {}
        self.all_data_set_jsons = {}
        self.all_pipelines = {}
        self.all_triggers = {}
        self.all_trigger_jsons = {}

        self.credentials_store_per_ir = {}

        self.source_data_sets_per_type = {}
        self.target_data_sets = {}

    def check_folders(self):
        for folder in [
            self.shir_folder, self.linked_service_folder,
            self.data_set_folder, self.pipeline_folder,
            self.trigger_folder
        ]:
            if not path.exists(folder):
                makedirs(folder)

    def get_configs(self):
        if not self.all_config_jsons:
            for config_file_name in listdir(self.config_folder):

                if not config_file_name.endswith(".json"):
                    continue

                file_path = f"{self.config_folder}/{config_file_name}"

                with open(file_path, "r") as json_file:
                    self.all_config_jsons.append(json.load(json_file))

        for config_json in self.all_config_jsons:
            yield config_json

    def find_self_hosted_integration_runtimes(self):
        for config in self.get_configs():
            if "self_hosted_integration_runtime" in config:
                shir_name = config["self_hosted_integration_runtime"]
                if shir_name not in self.all_self_hosted_integration_runtimes:
                    self.all_self_hosted_integration_runtimes[shir_name] = {
                        "name": shir_name,
                        "properties": {
                            "type": "SelfHosted"
                        }
                    }

    def find_all_connections(self):
        # Find all unique instances of connection and authentication types

        if not self.all_linked_services:
            for config in self.get_configs():
                conn = self.connection_types[config["connection"]]
                auth = config["authentication"]
                ir_name = config.get(
                    "self_hosted_integration_runtime",
                    self.base_integration_runtime)

                source_ls = conn.get_source_linked_service(auth)["name"]
                self.all_linked_services.add((source_ls, ir_name))

                for _, source_dataset in conn.source_data_sets.items():
                    self.all_data_sets.add(
                        (source_ls, ir_name, source_dataset["name"]))

                target_ls = conn.target_linked_service["name"]
                self.all_linked_services.add((target_ls, ir_name))

                for _, target_dataset in conn.target_data_sets.items():
                    self.all_data_sets.add(
                        (target_ls, ir_name, target_dataset["name"]))

                if conn.config_linked_service:
                    config_ls = conn.config_linked_service["name"]
                    self.all_linked_services.add((config_ls, ir_name))

                    for _, config_dataset in conn.config_data_sets.items():
                        self.all_data_sets.add(
                            (config_ls, ir_name, config_dataset["name"]))

            # Add the credentials store, if required
            add_credentials_store = set()
            for ls_name, ir_name in self.all_linked_services:
                base_json = all_linked_services[ls_name]

                for _, v in base_json["properties"]["typeProperties"].items():
                    if not isinstance(v, dict):
                        continue
                    if v.get("type") == "AzureKeyVaultSecret":
                        add_credentials_store.add(ir_name)

            # Add this outside the loop so the set doesn't change size
            for name in add_credentials_store:
                self.all_linked_services.add(("Credentials Store", name))
                self.credentials_store_per_ir[name] = \
                    self.create_linked_service_name(
                        "Credentials Store", name)

    @staticmethod
    def get_file_path(folder_path, json_to_write):
        return f"{folder_path}/{json_to_write['name']}.json"
    
    def get_current_json(self, folder_path, json_to_write):
        file_path = self.get_file_path(folder_path, json_to_write)

        if not path.isfile(file_path):
            return

        with open(file_path) as json_file:
            return json.load(json_file)

    def create_linked_service_name(self, base_name, integration_runtime_name):
        if integration_runtime_name == self.base_integration_runtime:
            return base_name
        else:
            return f"{base_name}{integration_runtime_name}"

    linked_service_preservation = (
        ("AzureBlobFS", "url"),
        ("AzureKeyVault", "baseUrl")
    )

    def create_linked_service(self, base_json, integration_runtime_name):
        new_ls_json = deepcopy(base_json)
        if integration_runtime_name != self.base_integration_runtime:
            new_ls_json["name"] = \
                self.create_linked_service_name(base_json["name"],
                                                integration_runtime_name)
            new_ls_json["properties"]["connectVia"] = {
                "referenceName": integration_runtime_name,
                "type": "IntegrationRuntimeReference"
            }

            for _, v in new_ls_json["properties"]["typeProperties"].items():
                if not isinstance(v, dict):
                    continue
                if v.get("type") == "AzureKeyVaultSecret":
                    v["store"]["referenceName"] = \
                        self.credentials_store_per_ir[integration_runtime_name]

        curr_file = self.get_current_json(
            self.linked_service_folder, new_ls_json)
        if curr_file and not self.overwrite:
            for ls_type, replace_str in self.linked_service_preservation:
                if curr_file["properties"]["type"] == ls_type:
                    new_ls_json["properties"]["typeProperties"][
                        replace_str
                    ] = curr_file["properties"]["typeProperties"][
                        replace_str
                    ]

        return new_ls_json

    def create_dataset_name(self, linked_service_name, data_set_template_name):

        ds_name = data_set_template_name
        if ds_name.lower().startswith(linked_service_name.lower()):
            ds_name = ds_name[len(linked_service_name):]

        prefix = 0
        while ds_name[:prefix + 1] == linked_service_name[:prefix + 1]:
            prefix += 1
        ds_name = ds_name[prefix:]

        return sub("[^0-9a-zA-Z_]+", "", linked_service_name + ds_name)

    def find_all_linked_services(self):
        # Find only the required linked services
        self.find_all_connections()

        for ls_name, ir_name in self.all_linked_services:

            template_linked_service = all_linked_services[ls_name]

            ls_json = self.create_linked_service(
                template_linked_service, ir_name)
            if ls_json["name"] not in self.all_linked_service_jsons:
                self.all_linked_service_jsons[ls_json["name"]] = ls_json

    def find_all_data_sets(self):
        for ls_name, ir_name, ds_name in self.all_data_sets:

            linked_service_json = self.all_linked_service_jsons[
                self.create_linked_service_name(
                    ls_name, ir_name)
            ]
            data_set_template = all_data_sets[ds_name]

            ds_id = (linked_service_json["name"], data_set_template["name"])
            if ds_id not in self.all_data_set_jsons:

                data_set_json = deepcopy(data_set_template)

                data_set_json["name"] = \
                    self.create_dataset_name(
                        linked_service_json["name"], ds_name)

                parameters = linked_service_json["properties"].get(
                    "parameters", {})

                # Add linked service definition, including required parameters
                data_set_json["properties"]["linkedServiceName"] = {
                    "referenceName": linked_service_json["name"],
                    "type": "LinkedServiceReference",
                }
                if parameters:
                    data_set_json["properties"]["linkedServiceName"]["parameters"] = {
                        param: {
                            "value": f"@dataset().{param}",
                            "type": "Expression"
                        }
                        for param in parameters
                    }
                # Add linked service parameters to data set parameters
                for param, val in parameters.items():
                    data_set_json["properties"]["parameters"][param] = val

                self.all_data_set_jsons[ds_id] = data_set_json

    def generate_pipelines(self):

        for config in self.get_configs():
            conn = config["connection"]
            auth = config["authentication"]
            ir_name = config.get(
                "self_hosted_integration_runtime",
                self.base_integration_runtime)

            # Get the required pipeline type for this configuation
            pipeline_class = self.connection_types[conn]

            # Add the schedule if we don't know about it already
            schedule_details = \
                config.get("schedule", pipeline_class.default_schedule)
            schedule_id = create_schedule_id(schedule_details)
            if schedule_id not in self.all_triggers:
                self.all_triggers[schedule_id] = []

            # Take the linked services and add the required data sets
            source_linked_service_name = \
                self.create_linked_service_name(
                    pipeline_class.get_source_linked_service(auth)["name"],
                    ir_name)
            target_linked_service_name = \
                self.create_linked_service_name(
                    pipeline_class.target_linked_service["name"],
                    ir_name)

            pipeline_datasets = {
                **{
                    data_set_id: self.all_data_set_jsons[(
                        source_linked_service_name, data_set_template["name"])]
                    for data_set_id, data_set_template
                    in pipeline_class.source_data_sets.items()
                },
                **{
                    data_set_id: self.all_data_set_jsons[(
                        target_linked_service_name, data_set_template["name"])]
                    for data_set_id, data_set_template
                    in pipeline_class.target_data_sets.items()
                }
            }

            # Add a configuration linked service, separate to source and
            # target, if required
            if pipeline_class.config_linked_service:
                config_linked_service_name = self.create_linked_service_name(
                    pipeline_class.config_linked_service["name"], ir_name)
                pipeline_datasets.update({
                    data_set_id: self.all_data_set_jsons[(
                        config_linked_service_name, data_set_template["name"])]
                    for data_set_id, data_set_template
                    in pipeline_class.config_data_sets.items()
                })

            # Create a pipeline per table
            for table_definition in config["tables"]:
                pipeline_obj = pipeline_class(
                    config["name"], config["authentication"],
                    config["config"], table_definition, pipeline_datasets
                )
                pipeline_obj.generate_pipeline()
                self.all_pipelines[pipeline_obj.pipeline_json["name"]] = \
                    pipeline_obj.pipeline_json
                self.all_triggers[schedule_id].append(
                    pipeline_obj.pipeline_json["name"])

    def generate_triggers(self):

        for schedule_id, pipelines in self.all_triggers.items():
            self.all_trigger_jsons[schedule_id] = {
                "name": trigger_name(*schedule_id),
                "properties": {
                    "annotations": [],
                    "runtimeState": "Started",
                    "pipelines": [
                        {
                            "pipelineReference": {
                                "referenceName": pipeline,
                                "type": "PipelineReference"
                            }
                        }
                        for pipeline in sorted(pipelines)
                    ],
                    "type": "ScheduleTrigger",
                    "typeProperties": {
                        "recurrence": create_recurrence_object(*schedule_id)
                    }
                }
            }

    @staticmethod
    def add_defaults(objects_json):
        for _, v in objects_json.items():
            v["properties"]["annotations"] = list(set(
                default_annotations + 
                v["properties"].get("annotations", [])
            ))
        
            if v.get("type", "").split("/")[-1] == "pipelines":
                if not v["properties"]["parameters"]:
                    del v["properties"]["parameters"]
                if not v["properties"]["variables"]:
                    del v["properties"]["variables"]

    def create_all_jsons(self):
        self.find_self_hosted_integration_runtimes()

        self.find_all_linked_services()
        self.add_defaults(self.all_linked_service_jsons)

        self.find_all_data_sets()
        self.add_defaults(self.all_data_set_jsons)

        self.generate_pipelines()
        self.add_defaults(self.all_pipelines)

        self.generate_triggers()
        self.add_defaults(self.all_trigger_jsons)

    def write_json(self, folder_path, json_to_write):
        file_path = self.get_file_path(folder_path, json_to_write)
        with open(file_path, "w") as json_file:
            json.dump(json_to_write, json_file, indent=4)

    @staticmethod
    def clean_unused_jsons(folder_path, generated_jsons):
        generated_json_names = [
            f"{json['name']}.json"
            for _, json in generated_jsons.items()
        ]
        all_files = [
            f 
            for f in listdir(folder_path) 
            if path.isfile(path.join(folder_path, f)) and f.endswith(".json")
        ]
        for file in all_files:
            if file not in generated_json_names:
                remove(f"{folder_path}/{file}")

    def create_all(self):
        self.create_all_jsons()

        for _, shir_json in self.all_self_hosted_integration_runtimes.items():
            self.write_json(self.shir_folder, shir_json)

        for _, ls_json in self.all_linked_service_jsons.items():
            self.write_json(self.linked_service_folder, ls_json) 

        for _, data_set_json in self.all_data_set_jsons.items():
            self.write_json(self.data_set_folder, data_set_json)

        for _, pipeline_json in self.all_pipelines.items():
            self.write_json(self.pipeline_folder, pipeline_json)

        for _, trigger_json in self.all_trigger_jsons.items():
            self.write_json(self.trigger_folder, trigger_json)

        if self.remove_not_generated:
            self.clean_unused_jsons(self.shir_folder, self.all_self_hosted_integration_runtimes)
            self.clean_unused_jsons(self.linked_service_folder, self.all_linked_service_jsons)
            self.clean_unused_jsons(self.data_set_folder, self.all_data_set_jsons)
            self.clean_unused_jsons(self.pipeline_folder, self.all_pipelines)
            self.clean_unused_jsons(self.trigger_folder, self.all_trigger_jsons)



from .config_table import config_table
from .data_lake import data_lake_folder, data_lake_file
from .ftp import ftp_folder, ftp_file
from .sftp import sftp_folder, sftp_file

all_data_sets = {
    ds["name"]: ds
    for ds in [
        config_table,
        data_lake_folder, data_lake_file,
        ftp_folder, ftp_file,
        sftp_folder, sftp_file
    ]
}



from .config_table_storage import config_table_storage
from .credentials_store import credentials_store
from .data_lake import data_lake
from .ftp import ftp_basic_key_vault
from .sftp import sftp_basic_key_vault

all_linked_services = {
    ls["name"]: ls
    for ls in [
        config_table_storage, credentials_store, data_lake,
        ftp_basic_key_vault, sftp_basic_key_vault
    ]
}



def get_data_lake_files(policy, data_set_name, container, path):
    return {
        "name": f"List {container}/{path} files".replace("/", "-"),
        "type": "GetMetadata",
        "dependsOn": [],
        "policy": policy,
        "userProperties": [],
        "typeProperties": {
            "dataset": {
                "referenceName": data_set_name,
                "type": "DatasetReference",
                "parameters": {
                    "Container": container,
                    "FolderPath": path,
                }
            }
        },
        "fieldList": [
            "childItems"
        ],
        "storeSettings": {
            "type": "AzureBlobFSReadSettings",
            "recursive": True,
            "enablePartitionDiscovery": False
        },
        "formatSettings": {
            "type": "BinaryReadSettings"
        }
    }



ftp_file = {
    "name": "FTPFile",
    "properties": {
        "linkedServiceName": {},
        "parameters": {
            "FolderPath": {
                "type": "string",
                "defaultValue": "/"
            },
            "FileName": {
                "type": "string"
            }
        },
        "annotations": [],
        "type": "Binary",
        "typeProperties": {
            "location": {
                "type": "FtpServerLocation",
                "FileName": {
                    "value": "@dataset().FileName",
                    "type": "Expression"
                },
                "FolderPath": {
                    "value": "@dataset().FolderPath",
                    "type": "Expression"
                }
            }
        }
    },
    "type": "Microsoft.DataFactory/factories/datasets"
}
ftp_folder = {
    "name": "FTPFolder",
    "properties": {
        "linkedServiceName": {},
        "parameters": {
            "FolderPath": {
                "type": "string",
                "defaultValue": "/"
            }
        },
        "annotations": [],
        "type": "Binary",
        "typeProperties": {
            "location": {
                "type": "FtpServerLocation",
                "folderPath": {
                    "value": "@dataset().FolderPath",
                    "type": "Expression"
                }
            }
        }
    },
    "type": "Microsoft.DataFactory/factories/datasets"
}



credentials_store = {
	"name": "Credentials Store",
	"type": "Microsoft.DataFactory/factories/linkedservices",
	"properties": {
		"annotations": [],
		"type": "AzureKeyVault",
		"typeProperties": {
			"baseUrl": "https://examplekeyvaultname.vault.azure.net/"
		}
	}
}


from os import path, makedirs
import typer

from .initialise import CreateDataFactoryObjects

app = typer.Typer(name="ingeniiadfg")

@app.callback()
def callback():
    """
    Ingenii Azure Data Factory Generator
    """

@app.command()
def generate(
    configs_folder: str,
    target_folder: str = typer.Option(
        ".",
        help="The root path to write the object .JSON files. Defaults to the folder the command is run in",
    ),
    overwrite: bool = typer.Option(
        False,
        help="Overwrites objects completely, ignoring what's been set so far",
    ),
    remove_not_generated: bool = typer.Option(
        False,
        help="Delete any other objects not generated by this command",
    )
        ):
    """
    Generate Data Factory objects from files in the CONFIGS_FOLDER
    """

    if not path.exists(configs_folder):
        raise Exception(
            f"Provided config files folder `{configs_folder}`"
            " does not exist!")
    if target_folder != "." and not path.exists(target_folder):
        makedirs(target_folder)

    initialisation_obj = CreateDataFactoryObjects(
        configs_folder, target_folder,
        overwrite=overwrite,
        remove_not_generated=remove_not_generated
    )
    initialisation_obj.create_all()



sftp_basic_key_vault = {
    "name": "SFTPBasic",
    "type": "Microsoft.DataFactory/factories/linkedservices",
    "properties": {
        "type": "Sftp",
        "parameters": {
            "Host": {
                "type": "String"
            },
            "KeyVaultSecretName": {
                "type": "String"
            },
            "UserName": {
                "type": "String"
            },
            "Port": {
                "type": "Int",
                "defaultValue": 22
            },
        },
        "annotations": [],
        "typeProperties": {
            "host": "@{linkedService().Host}",
            "port": "@linkedService().Port",
            "skipHostKeyValidation": True,
            "authenticationType": "Basic",
            "userName": "@{linkedService().UserName}",
            "password": {
                "type": "AzureKeyVaultSecret",
                "store": {
                    "referenceName": "Credentials Store",
                    "type": "LinkedServiceReference"
                },
                "secretName": "@{linkedService().KeyVaultSecretName}"
            }
        }
    }
}



data_lake_folder = {
    "name": "DataLakeFolder",
    "properties": {
        "linkedServiceName": {},
        "parameters": {
            "Container": {
                "type": "string"
            },
            "FolderPath": {
                "type": "string"
            }
        },
        "annotations": [],
        "type": "Binary",
        "typeProperties": {
            "location": {
                "type": "AzureBlobFSLocation",
                "folderPath": {
                    "value": "@dataset().FolderPath",
                    "type": "Expression"
                },
                "fileSystem": {
                    "value": "@dataset().Container",
                    "type": "Expression"
                }
            }
        }
    },
    "type": "Microsoft.DataFactory/factories/datasets"
}
data_lake_file = {
    "name": "DataLakeFile",
    "properties": {
        "linkedServiceName": {},
        "parameters": {
            "Container": {
                "type": "string"
            },
            "FolderPath": {
                "type": "string"
            },
            "FileName": {
                "type": "string"
            }
        },
        "annotations": [],
        "type": "Binary",
        "typeProperties": {
            "location": {
                "type": "AzureBlobFSLocation",
                "fileName": {
                    "value": "@dataset().FileName",
                    "type": "Expression"
                },
                "folderPath": {
                    "value": "@dataset().FolderPath",
                    "type": "Expression"
                },
                "fileSystem": {
                    "value": "@dataset().Container",
                    "type": "Expression"
                }
            }
        }
    },
    "type": "Microsoft.DataFactory/factories/datasets"
}



data_lake = {
    "name": "DataLake",
    "type": "Microsoft.DataFactory/factories/linkedservices",
    "properties": {
        "annotations": [],
        "type": "AzureBlobFS",
        "typeProperties": {
            "url": "https://exampledatalakename.dfs.core.windows.net"
        }
    }
}



schedule = {
    "name": "schedule",
    "properties": {
        "annotations": [],
        "runtimeState": "Started",
        "pipelines": [],
        "type": "ScheduleTrigger",
        "typeProperties": {
            "recurrence": {}
        }
    }
}



config_table_storage = {
    "name": "ConfigTableStorage",
    "type": "Microsoft.DataFactory/factories/linkedservices",
    "properties": {
        "annotations": [],
        "type": "AzureTableStorage",
        "typeProperties": {
            "sasUri": {
                "type": "AzureKeyVaultSecret",
                "store": {
                    "referenceName": "Credentials Store",
                    "type": "LinkedServiceReference"
                },
                "secretName": "datalake-table-storage-sas-uri"
            }
        }
    }
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ingeniiadfg',
 'ingeniiadfg.activities',
 'ingeniiadfg.templates',
 'ingeniiadfg.templates.dataset',
 'ingeniiadfg.templates.integration_runtime',
 'ingeniiadfg.templates.linked_service',
 'ingeniiadfg.templates.trigger']

package_data = \
{'': ['*']}

install_requires = \
['deepdiff==5.5.0', 'typer[all]>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['ingeniiadfg = ingeniiadfg.main:app']}

setup_kwargs = {
    'name': 'ingeniiadfg',
    'version': '0.1.5',
    'description': 'Generate Azure Data Factory objects from configuration',
    'long_description': "# Ingenii Azure Data Factory Generator\nPython based generator to create Azure Data Factory pipelines from configurations.\n\nThis package integrates easily with the [Ingenii Azure Data Platform](https://github.com/ingenii-solutions/azure-data-platform), but this package can be used independently as long as some resouces are created ahead of time. These are detailed in the [Usage documentation](docs/user/Usage.md).\n\n* Current Version: 0.1.5\n\n## Package installation\n\nInstall the package [using pip](https://pip.pypa.io/en/stable/user_guide/) with \n```\npip install ingeniiadfg\n```\nor, for a particular version\n```\npip install ingeniiadfg==0.1.5\n```\nAlternatively, add it to your repository's current `requirements.txt` file. \n\n## Using the package\n\nFor details on using the package please refer to the [Azure Data Factory Usage documentation](docs/user/Usage.md).\n\n## Example CI/CD\n\nFor deploying into a Data Factory that is not integrated with a repository, also known as 'live' mode, we have included some example CI/CD pipelines in the `CICD` folder. These are in the format to be read by Azure Pipelines. Feel free to use these yourself or for inspiration in creating your own pipelines. \n\n## Version History\n\n* `0.1.5`: Move to a CLI package, and add several fixes \n* `0.1.4`: Add object annotations to track what is managed by this package \n* `0.1.3`: Extend schedule to handle when only the hours of the dayt are specified \n* `0.1.2`: Change the name of the secret name for the SAS URI to access the config tables\n* `0.1.1`: Add schedule generation from configuration, many more tests\n* `0.1.0`: Initial package, FTP/SFTP connections with basic authentication\n",
    'author': 'Greg Atkins',
    'author_email': 'greg@ingenii.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from .base import DataFactoryPipeline
from .templates.linked_service import config_table_storage, data_lake, \
    ftp_basic_key_vault, sftp_basic_key_vault
from .templates.dataset import config_table, data_lake_folder, ftp_folder, \
    ftp_file, sftp_folder, sftp_file


class FTPBasePipeline(DataFactoryPipeline):

    target_linked_service = data_lake
    target_data_sets = {
        "target_folder": data_lake_folder
    }

    config_linked_service = config_table_storage
    config_data_sets = {
        "config_table": config_table
    }

    default_schedule = {
        "type": "day",
        "time": "06:00"
    }

    required_table_parameters = ["name", "path"]
    # TODO: implement prefix and suffix checks
    # optional_table_parameters = ["zipped", "prefix", "suffix"]

    def __init__(self, data_provider, authentication,
                 config, table_definition, data_sets):
        self.data_provider = data_provider
        self.authentication = authentication
        self.config = config
        self.table_definition = table_definition
        self.data_sets = data_sets

        self.table_storage_partition_key = \
            f"{data_provider}-{table_definition['name']}"

        self.data_lake_path = self.data_provider + \
            "/" + self.table_definition["name"]

        self.source_parameters = {
            "Host": self.config["host"],
            "UserName": self.config["username"],
            "KeyVaultSecretName": self.config["key_vault_secret_name"],
            "FolderPath": self.table_definition["path"]
        }
        if self.config.get("custom_port"):
            self.source_parameters["Port"] = self.config["custom_port"]

        super(FTPBasePipeline, self).__init__(
            self.data_lake_path.replace("/", "-"),
            variables={
                "SFTPKnownFiles": {
                    "type": "Array"
                }
            }
        )

    @staticmethod
    def handle_path(path_str):
        if path_str == "/":
            return "root path"
        else:
            return path_str.strip("/").replace("/", "-")

    def list_source_files(self):
        return {
            "name": "List files at " +
                    self.handle_path(self.table_definition['path']),
            "type": "GetMetadata",
            "dependsOn": [],
            "policy": self.default_policy,
            "userProperties": [],
            "typeProperties": {
                "dataset": self.create_pipeline_dataset_reference(
                    self.data_sets["source_folder"], self.source_parameters),
                "fieldList": ["childItems"],
                "storeSettings": {
                    "type": self.name.title() + "ReadSettings",
                    "recursive": True,
                    "enablePartitionDiscovery": False
                },
                "formatSettings": {"type": "BinaryReadSettings"}
                }
            }

    def list_known_files(self):
        return {
            "name": "Get known files",
            "type": "Lookup",
            "dependsOn": [],
            "policy": self.default_policy,
            "userProperties": [],
            "typeProperties": {
                "source": {
                    "type": "AzureTableSource",
                    "azureTableSourceQuery": {
                        "value": f"PartitionKey eq "
                                 f"'{self.table_storage_partition_key}'",
                        "type": "Expression"
                    },
                    "azureTableSourceIgnoreTableNotFound": True
                },
                "dataset": {
                    "referenceName": self.data_sets["config_table"]["name"],
                    "type": "DatasetReference",
                    "parameters": {
                        "TableName": "SFTPKnownFiles"
                    }
                },
                "firstRowOnly": False
            }
        }

    def generate_known_files_array(self, known_files_activity):
        return {
            "name": "Each known file",
            "type": "ForEach",
            "userProperties": [],
            "typeProperties": {
                "items": {
                    "value": "@activity('" +
                             known_files_activity['name'] +
                             "').output.value",
                    "type": "Expression"
                },
                "activities": [
                    {
                        "name": "Add file name",
                        "type": "AppendVariable",
                        "dependsOn": [],
                        "userProperties": [],
                        "typeProperties": {
                            "variableName": "SFTPKnownFiles",
                            "value": {
                                "value": "@item().RowKey",
                                "type": "Expression"
                            }
                        }
                    }
                ]
            }
        }

    def filter_new_files(self, source_files_activity):
        return {
            "name": "Find new files",
            "type": "Filter",
            "userProperties": [],
            "typeProperties": {
                "items": {
                    "value": "@activity('" +
                             source_files_activity['name'] +
                             "').output.childItems",
                    "type": "Expression"
                },
                "condition": {
                    "value": "@not(contains(variables('SFTPKnownFiles'), " +
                             "item().name))",
                    "type": "Expression"
                }
            }
        }

    def move_new_files(self, new_files_activity):
        sub_activities = [
            {
                "name": "Move file",
                "type": "Copy",
                "dependsOn": [],
                "policy": self.default_policy,
                "userProperties": [],
                "typeProperties": {
                    "source": {
                        "type": "BinarySource",
                        "storeSettings": self.source_store_settings,
                        "formatSettings": {
                            "type": "BinaryReadSettings"
                        }
                    },
                    "sink": {
                        "type": "BinarySink",
                        "storeSettings": {
                            "type": "AzureBlobFSWriteSettings"
                        }
                    },
                    "enableStaging": False
                },
                "inputs": [self.create_pipeline_dataset_reference(
                    self.data_sets["source_file"],
                    {
                        **self.source_parameters,
                        "FileName": {
                            "value": "@item().name",
                            "type": "Expression"
                        }
                     }
                )],
                "outputs": [self.create_pipeline_dataset_reference(
                    self.data_sets["target_folder"],
                    {
                        "Container": "raw",
                        "FolderPath": self.data_lake_path
                    }
                )]
            },
            {
                "name": "Create new known file entry",
                "type": "Copy",
                "dependsOn": [
                    {
                        "activity": "Move file",
                        "dependencyConditions": [
                            "Succeeded"
                        ]
                    }
                ],
                "policy": self.default_policy,
                "userProperties": [],
                "typeProperties": {
                    "source": {
                        "type": "AzureTableSource",
                        "additionalColumns": [
                            {
                                "name": "Row",
                                "value": {
                                    "value": "@item().name",
                                    "type": "Expression"
                                }
                            },
                            {
                                "name": "DateMoved",
                                "value": {
                                    "value": "@utcnow()",
                                    "type": "Expression"
                                }
                            }
                        ],
                        "azureTableSourceQuery": {
                            "value": "PartitionKey eq '1'",
                            "type": "Expression"
                        },
                        "azureTableSourceIgnoreTableNotFound": False
                    },
                    "sink": {
                        "type": "AzureTableSink",
                        "azureTableInsertType": "merge",
                        "azureTableDefaultPartitionKeyValue": {
                            "value": self.table_storage_partition_key,
                            "type": "Expression"
                        },
                        "azureTableRowKeyName": {
                            "value": "Row",
                            "type": "Expression"
                        },
                        "writeBatchSize": 10000
                    },
                    "enableStaging": False,
                    "translator": {
                        "type": "TabularTranslator",
                        "typeConversion": True,
                        "typeConversionSettings": {
                            "allowDataTruncation": True,
                            "treatBooleanAsNumber": False
                        }
                    }
                },
                "inputs": [
                    self.create_pipeline_dataset_reference(
                        self.data_sets["config_table"],
                        parameters={"TableName": "Select1"}
                    )
                ],
                "outputs": [
                    self.create_pipeline_dataset_reference(
                        self.data_sets["config_table"],
                        parameters={"TableName": "SFTPKnownFiles"}
                    )
                ]
            }
        ]
        return {
            "name": "For each new file",
            "type": "ForEach",
            "userProperties": [],
            "typeProperties": {
                "items": {
                    "value": "@activity('" +
                             new_files_activity['name'] +
                             "').output.value",
                    "type": "Expression"
                },
                "activities": sub_activities
            }
        }

    def generate_pipeline(self):

        # --

        find_known_files_raw = self.list_known_files()
        self.add_activity(find_known_files_raw)

        known_files_array = self.generate_known_files_array(
            find_known_files_raw)
        self.add_activity(known_files_array, depends_on=[find_known_files_raw])

        # --

        source_files = self.list_source_files()
        self.add_activity(source_files)

        only_new_files = self.filter_new_files(source_files)
        self.add_activity(only_new_files, depends_on=[
                          known_files_array, source_files])

        # --

        move_files = self.move_new_files(only_new_files)
        self.add_activity(move_files, depends_on=[only_new_files])


class FTPPipeline(FTPBasePipeline):

    name = "ftp"
    authentications = {
        "basic": {
            "reqired_config": [
                "host", "username", "key_vault_name", "key_vault_secret"],
            "linked_service": ftp_basic_key_vault
        }
    }
    source_data_sets = {
        "source_folder": ftp_folder,
        "source_file": ftp_file
    }

    source_store_settings = {
        "type": "FtpReadSettings",
        "recursive": False,
        "useBinaryTransfer": True,
        "deleteFilesAfterCompletion": False
    }

    def __init__(self, data_provider, authentication,
                 config, table_definition, data_sets):
        super(FTPPipeline, self).__init__(
            data_provider, authentication,
            config, table_definition, data_sets)


class SFTPPipeline(FTPBasePipeline):

    name = "sftp"
    authentications = {
        "basic": {
            "required_config": [
                "host", "username", "key_vault_name", "key_vault_secret"],
            "linked_service": sftp_basic_key_vault
        }
    }
    source_data_sets = {
        "source_folder": sftp_folder,
        "source_file": sftp_file
    }

    source_store_settings = {
        "type": "SftpReadSettings",
        "recursive": False,
        "deleteFilesAfterCompletion": False
    }

    def __init__(self, data_provider, authentication,
                 config, table_definition, data_sets):
        super(SFTPPipeline, self).__init__(
            data_provider, authentication,
            config, table_definition, data_sets)



from abc import ABC, abstractmethod
from typing import List


class DataFactoryPipeline(ABC):

    # The general name of the connection type
    name = None

    # Dictionary where the keys are the names of the possible authentication
    # options e.g. 'basic', 'token'. The values are a sub-dictionary with 2
    # keys, 'required', and 'optional', and those values are a list of the
    # corresponding parameters required or optional for this connection type
    authentications = {}

    # A list of strings giving the names of the required and optional
    # parameters to define for each table
    required_table_parameters = []
    optional_table_parameters = []

    default_policy = {
        "timeout": "0.00:01:00",
        "retry": 3,
        "retryIntervalInSeconds": 30,
        "secureOutput": False,
        "secureInput": False
    }

    data_provider = None
    authentication = None
    config = {}
    table_definition = {}
    data_sets = {}
    source_store_settings = {}

    source_data_sets = {}
    target_linked_service = None
    target_data_sets = {}
    config_linked_service = None
    config_data_sets = {}

    default_schedule = {}

    @classmethod
    def is_valid_authentication(cls, authentication_method: str) -> bool:
        return authentication_method in cls.authentications

    @classmethod
    def all_authentications(cls) -> List[str]:
        return list(cls.authentications.keys())

    @staticmethod
    def _validate_config(config: dict, required_parameters: List[str],
                         optional_parameters: List[str]) -> None:
        """
        Validate any config given the lists of parameters
        """
        errors = []

        missing_parameters = [
            rp for rp in required_parameters
            if rp not in config
        ]
        if missing_parameters:
            errors.append(f"Fields missing from config: {missing_parameters}")

        unknown_parameters = [
            c for c in config
            if c not in required_parameters + optional_parameters
        ]
        if unknown_parameters:
            errors.append(f"Unknown fields in config: {unknown_parameters}")

        if errors:
            raise Exception(errors)

    def __init__(self, name, parameters={}, variables={}, annotations=[]):
        self.pipeline_json = {
            "name": name,
            "type": "Microsoft.DataFactory/factories/pipelines",
            "properties": {
                "activities": [],
                "parameters": parameters,
                "variables": variables,
                "annotations": annotations
            }
        }
        if not self.source_store_settings:
            raise Exception(
                f"source_store_settings not set for data type {self.name}!")

    @classmethod
    def get_source_linked_service(cls, connection_name):
        return cls.authentications[connection_name]["linked_service"]

    def add_activity(self, activity_json, depends_on=[]):
        self.pipeline_json["properties"]["activities"].append({
            **activity_json,
            "dependsOn": [
                {
                    "activity": activity["name"],
                    "dependencyConditions": ["Succeeded"]
                }
                for activity in depends_on
                ]
            })

    def create_pipeline_dataset_reference(self, data_set_json, parameters={}):
        missing_parameters = [
            parameter_name
            for parameter_name, val
            in data_set_json["properties"].get("parameters", {}).items()
            if parameter_name not in parameters
            and val.get("defaultValue") is None
        ]
        if missing_parameters:
            raise Exception(f"Missing parameters when accessing dataset "
                            f"{data_set_json['name']}: {missing_parameters}")
        return {
            "referenceName": data_set_json["name"],
            "type": "DatasetReference",
            "parameters": parameters
        }

    def list_target_files(
                self, container, path,
                policy_ovverride={}):
        return {
            "name": f"List {container}/{path} files".replace("/", "-"),
            "type": "GetMetadata",
            "dependsOn": [],
            "policy": {**self.default_policy, **policy_ovverride},
            "userProperties": [],
            "typeProperties": {
                "dataset": self.create_pipeline_dataset_reference(
                    self.data_sets["target_folder"], {
                        "Container": container,
                        "FolderPath": path
                    })
            },
            "fieldList": [
                "childItems"
            ],
            "storeSettings": {
                "type": "AzureBlobFSReadSettings",
                "recursive": True,
                "enablePartitionDiscovery": False
            },
            "formatSettings": {
                "type": "BinaryReadSettings"
            }
        }

    @abstractmethod
    def generate_pipeline(self):
        ...



config_table = {
    "name": "ConfigTable",
    "properties": {
        "linkedServiceName": {
            "referenceName": "AzureConfigTableStorage",
            "type": "LinkedServiceReference"
        },
        "parameters": {
            "TableName": {
                "type": "string"
            }
        },
        "annotations": [],
        "type": "AzureTable",
        "schema": [],
        "typeProperties": {
            "tableName": {
                "value": "@dataset().TableName",
                "type": "Expression"
            }
        }
    },
    "type": "Microsoft.DataFactory/factories/datasets"
}
